@@ -13,13 +13,15 @@
 // License for the specific language governing permissions and limitations
 // under the License.
 
-//! Stored program manipulation and interactive editor.
+//! Stored program manipulation.
 
 use crate::console::Console;
 use async_trait::async_trait;
 use endbasic_core::ast::{ArgSep, Expr, Value, VarType};
-use endbasic_core::eval::{CallableMetadata, CallableMetadataBuilder};
-use endbasic_core::exec::{self, Command, Machine};
+use endbasic_core::exec::Machine;
+use endbasic_core::syms::{
+    CallError, CallableMetadata, CallableMetadataBuilder, Command, CommandResult,
+};
 use std::cell::RefCell;
 use std::collections::{BTreeMap, HashMap};
 use std::fs::{self, File, OpenOptions};
@@ -28,6 +30,9 @@ use std::path::PathBuf;
 use std::rc::Rc;
 use std::str;
 
+/// Category string for all functions provided by this module.
+const CATEGORY: &str = "Stored program manipulation";
+
 /// Metadata of an entry in the store.
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct Metadata {
@@ -258,7 +263,7 @@ impl DelCommand {
         Rc::from(Self {
             metadata: CallableMetadataBuilder::new("DEL", VarType::Void)
                 .with_syntax("filename")
-                .with_category("Stored program manipulation")
+                .with_category(CATEGORY)
                 .with_description(
                     "Deletes the given program.
 The filename must be a string and must be a basename (no directory components).  The .BAS \
@@ -276,21 +281,21 @@ impl Command for DelCommand {
         &self.metadata
     }
 
-    async fn exec(
-        &self,
-        args: &[(Option<Expr>, ArgSep)],
-        machine: &mut Machine,
-    ) -> exec::Result<()> {
+    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> CommandResult {
         if args.len() != 1 {
-            return exec::new_usage_error("DEL requires a filename");
+            return Err(CallError::ArgumentError("DEL requires a filename".to_owned()));
         }
         let arg0 = args[0].0.as_ref().expect("Single argument must be present");
-        match arg0.eval(machine.get_vars(), machine.get_functions())? {
+        match arg0.eval(machine.get_mut_symbols())? {
             Value::Text(t) => {
                 let name = to_filename(t)?;
                 self.store.borrow_mut().delete(&name)?;
             }
-            _ => return exec::new_usage_error("DEL requires a string as the filename"),
+            _ => {
+                return Err(CallError::ArgumentError(
+                    "DEL requires a string as the filename".to_owned(),
+                ))
+            }
         }
         Ok(())
     }
@@ -309,7 +314,7 @@ impl DirCommand {
         Rc::from(Self {
             metadata: CallableMetadataBuilder::new("DIR", VarType::Void)
                 .with_syntax("")
-                .with_category("Stored program manipulation")
+                .with_category(CATEGORY)
                 .with_description("Displays the list of files on disk.")
                 .build(),
             console,
@@ -324,13 +329,9 @@ impl Command for DirCommand {
         &self.metadata
     }
 
-    async fn exec(
-        &self,
-        args: &[(Option<Expr>, ArgSep)],
-        _machine: &mut Machine,
-    ) -> exec::Result<()> {
+    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], _machine: &mut Machine) -> CommandResult {
         if !args.is_empty() {
-            return exec::new_usage_error("DIR takes no arguments");
+            return Err(CallError::ArgumentError("DIR takes no arguments".to_owned()));
         }
         show_dir(&*self.store.borrow(), &mut *self.console.borrow_mut())?;
         Ok(())
@@ -350,7 +351,7 @@ impl EditCommand {
         Rc::from(Self {
             metadata: CallableMetadataBuilder::new("EDIT", VarType::Void)
                 .with_syntax("")
-                .with_category("Stored program manipulation")
+                .with_category(CATEGORY)
                 .with_description("Interactively edits the stored program.")
                 .build(),
             console,
@@ -365,13 +366,9 @@ impl Command for EditCommand {
         &self.metadata
     }
 
-    async fn exec(
-        &self,
-        args: &[(Option<Expr>, ArgSep)],
-        _machine: &mut Machine,
-    ) -> exec::Result<()> {
+    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], _machine: &mut Machine) -> CommandResult {
         if !args.is_empty() {
-            return exec::new_usage_error("EDIT takes no arguments");
+            return Err(CallError::ArgumentError("EDIT takes no arguments".to_owned()));
         }
 
         let mut console = self.console.borrow_mut();
@@ -381,6 +378,58 @@ impl Command for EditCommand {
     }
 }
 
+/// The `LIST` command.
+pub struct ListCommand {
+    metadata: CallableMetadata,
+    console: Rc<RefCell<dyn Console>>,
+    program: Rc<RefCell<dyn Program>>,
+}
+
+impl ListCommand {
+    /// Creates a new `LIST` command that dumps the `program` to the `console`.
+    pub fn new(console: Rc<RefCell<dyn Console>>, program: Rc<RefCell<dyn Program>>) -> Rc<Self> {
+        Rc::from(Self {
+            metadata: CallableMetadataBuilder::new("LIST", VarType::Void)
+                .with_syntax("")
+                .with_category(CATEGORY)
+                .with_description("Prints the currently-loaded program.")
+                .build(),
+            console,
+            program,
+        })
+    }
+}
+
+#[async_trait(?Send)]
+impl Command for ListCommand {
+    fn metadata(&self) -> &CallableMetadata {
+        &self.metadata
+    }
+
+    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], _machine: &mut Machine) -> CommandResult {
+        if !args.is_empty() {
+            return Err(CallError::ArgumentError("LIST takes no arguments".to_owned()));
+        }
+        let program = self.program.borrow().text();
+        let program: Vec<&str> = program.lines().collect();
+        let digits = {
+            let mut digits = 0;
+            let mut count = program.len();
+            while count > 0 {
+                digits += 1;
+                count /= 10;
+            }
+            digits
+        };
+        let mut console = self.console.borrow_mut();
+        for (i, line) in program.into_iter().enumerate() {
+            let formatted = format!("{:digits$} | {}", i + 1, line, digits = digits);
+            console.print(&formatted)?;
+        }
+        Ok(())
+    }
+}
+
 /// The `LOAD` command.
 pub struct LoadCommand {
     metadata: CallableMetadata,
@@ -394,7 +443,7 @@ impl LoadCommand {
         Rc::from(Self {
             metadata: CallableMetadataBuilder::new("LOAD", VarType::Void)
                 .with_syntax("filename")
-                .with_category("Stored program manipulation")
+                .with_category(CATEGORY)
                 .with_description(
                     "Loads the given program.
 The filename must be a string and must be a basename (no directory components).  The .BAS \
@@ -413,23 +462,23 @@ impl Command for LoadCommand {
         &self.metadata
     }
 
-    async fn exec(
-        &self,
-        args: &[(Option<Expr>, ArgSep)],
-        machine: &mut Machine,
-    ) -> exec::Result<()> {
+    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> CommandResult {
         if args.len() != 1 {
-            return exec::new_usage_error("LOAD requires a filename");
+            return Err(CallError::ArgumentError("LOAD requires a filename".to_owned()));
         }
         let arg0 = args[0].0.as_ref().expect("Single argument must be present");
-        match arg0.eval(machine.get_vars(), machine.get_functions())? {
+        match arg0.eval(machine.get_mut_symbols())? {
             Value::Text(t) => {
                 let name = to_filename(t)?;
                 let content = self.store.borrow().get(&name)?;
                 self.program.borrow_mut().load(&content);
                 machine.clear();
             }
-            _ => return exec::new_usage_error("LOAD requires a string as the filename"),
+            _ => {
+                return Err(CallError::ArgumentError(
+                    "LOAD requires a string as the filename".to_owned(),
+                ))
+            }
         }
         Ok(())
     }
@@ -447,7 +496,7 @@ impl NewCommand {
         Rc::from(Self {
             metadata: CallableMetadataBuilder::new("NEW", VarType::Void)
                 .with_syntax("")
-                .with_category("Stored program manipulation")
+                .with_category(CATEGORY)
                 .with_description("Clears the stored program from memory.")
                 .build(),
             program,
@@ -461,13 +510,9 @@ impl Command for NewCommand {
         &self.metadata
     }
 
-    async fn exec(
-        &self,
-        args: &[(Option<Expr>, ArgSep)],
-        machine: &mut Machine,
-    ) -> exec::Result<()> {
+    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> CommandResult {
         if !args.is_empty() {
-            return exec::new_usage_error("NEW takes no arguments");
+            return Err(CallError::ArgumentError("NEW takes no arguments".to_owned()));
         }
         self.program.borrow_mut().load("");
         machine.clear();
@@ -490,7 +535,7 @@ impl RunCommand {
         Rc::from(Self {
             metadata: CallableMetadataBuilder::new("RUN", VarType::Void)
                 .with_syntax("")
-                .with_category("Stored program manipulation")
+                .with_category(CATEGORY)
                 .with_description(
                     "Runs the stored program.
 Note that the program runs in the context of the interpreter so it will pick up any variables \
@@ -509,16 +554,19 @@ impl Command for RunCommand {
         &self.metadata
     }
 
-    async fn exec(
-        &self,
-        args: &[(Option<Expr>, ArgSep)],
-        machine: &mut Machine,
-    ) -> exec::Result<()> {
+    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> CommandResult {
         if !args.is_empty() {
-            return exec::new_usage_error("RUN takes no arguments");
+            return Err(CallError::ArgumentError("RUN takes no arguments".to_owned()));
         }
         let program = self.program.borrow().text();
-        let stop_reason = machine.exec(&mut program.as_bytes()).await?;
+        let stop_reason = match machine.exec(&mut program.as_bytes()).await {
+            Ok(stop_reason) => stop_reason,
+            Err(e) => {
+                // TODO(jmmv): This conversion to an internal error is not great and is just a
+                // workaround for the mess that CallError currently is in the context of commands.
+                return Err(CallError::InternalError(format!("{}", e)));
+            }
+        };
         if stop_reason.as_exit_code() != 0 {
             self.console
                 .borrow_mut()
@@ -541,7 +589,7 @@ impl SaveCommand {
         Rc::from(Self {
             metadata: CallableMetadataBuilder::new("SAVE", VarType::Void)
                 .with_syntax("filename")
-                .with_category("Stored program manipulation")
+                .with_category(CATEGORY)
                 .with_description(
                     "Saves the current program in memory to the given filename.
 The filename must be a string and must be a basename (no directory components).  The .BAS \
@@ -560,22 +608,22 @@ impl Command for SaveCommand {
         &self.metadata
     }
 
-    async fn exec(
-        &self,
-        args: &[(Option<Expr>, ArgSep)],
-        machine: &mut Machine,
-    ) -> exec::Result<()> {
+    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> CommandResult {
         if args.len() != 1 {
-            return exec::new_usage_error("SAVE requires a filename");
+            return Err(CallError::ArgumentError("SAVE requires a filename".to_owned()));
         }
         let arg0 = args[0].0.as_ref().expect("Single argument must be present");
-        match arg0.eval(machine.get_vars(), machine.get_functions())? {
+        match arg0.eval(machine.get_mut_symbols())? {
             Value::Text(t) => {
                 let name = to_filename(t)?;
                 let content = self.program.borrow().text();
                 self.store.borrow_mut().put(&name, &content)?;
             }
-            _ => return exec::new_usage_error("SAVE requires a string as the filename"),
+            _ => {
+                return Err(CallError::ArgumentError(
+                    "SAVE requires a string as the filename".to_owned(),
+                ))
+            }
         }
         Ok(())
     }
@@ -592,6 +640,7 @@ pub fn add_all(
     machine.add_command(DelCommand::new(store.clone()));
     machine.add_command(DirCommand::new(console.clone(), store.clone()));
     machine.add_command(EditCommand::new(console.clone(), program.clone()));
+    machine.add_command(ListCommand::new(console.clone(), program.clone()));
     machine.add_command(LoadCommand::new(store.clone(), program.clone()));
     machine.add_command(NewCommand::new(program.clone()));
     machine.add_command(RunCommand::new(console, program.clone()));
@@ -822,6 +871,33 @@ mod tests {
         check_stmt_err("EDIT takes no arguments", "EDIT 1");
     }
 
+    #[test]
+    fn test_list_ok() {
+        Tester::default().run("LIST").check();
+
+        Tester::default()
+            .set_program("one\n\nthree\n")
+            .run("LIST")
+            .expect_prints(["1 | one", "2 | ", "3 | three"])
+            .expect_program("one\n\nthree\n")
+            .check();
+
+        Tester::default()
+            .set_program("a\nb\nc\nd\ne\nf\ng\nh\ni\nj\n")
+            .run("LIST")
+            .expect_prints([
+                " 1 | a", " 2 | b", " 3 | c", " 4 | d", " 5 | e", " 6 | f", " 7 | g", " 8 | h",
+                " 9 | i", "10 | j",
+            ])
+            .expect_program("a\nb\nc\nd\ne\nf\ng\nh\ni\nj\n")
+            .check();
+    }
+
+    #[test]
+    fn test_list_errors() {
+        check_stmt_err("LIST takes no arguments", "LIST 2");
+    }
+
     #[test]
     fn test_load_ok() {
         let content = "line 1\n\n  line 2\n";
